@@ -26,6 +26,7 @@ from functools import wraps
 from gettext import ngettext
 from optparse import OptionParser
 import os
+import subprocess
 import sys
 import time
 
@@ -84,108 +85,13 @@ def run_command(db, cmd, args):
 
 # Commands
 
-@command("open an the backend's interactive shell", aliases=('shell',))
+@command("open the backend's interactive shell", aliases=('shell',))
 def backend(db, args):
     parser = OptionParser(usage='''usage: %prog backend
 
 Run an interactive database session on the timebook database. Requires
 the sqlite3 command.''')
-    os.execvp('sqlite3', ('sqlite3', db.path))
-
-@command('display the current timesheet', aliases=('show',))
-def display(db, args):
-    parser = OptionParser(usage='''usage: %prog display [TIMESHEET]
-
-Display a given timesheet. If no timesheet is specified, show the
-current timesheet.''')
-    opts, args = parser.parse_args(args=args)
-    if args:
-        sheet = cmdutil.complete(dbutil.get_sheet_names(db), args[0],
-                                 'timesheet')
-    else:
-        sheet = dbutil.get_current_sheet(db)
-    print 'Timesheet %s:' % sheet
-    db.execute(u'''
-    select count(*) > 0 from entry where sheet = ?
-    ''', (sheet,))
-    if not db.fetchone()[0]:
-        print '(empty)'
-        return
-
-    displ_time = lambda t: time.strftime('%H:%M:%S', time.localtime(t))
-    displ_date = lambda t: time.strftime('%b %d, %Y',
-                                         time.localtime(t))
-    displ_total = lambda t: str(timedelta(seconds=t))
-    last_day = None
-    table = [['Day', 'Start      End', 'Duration', 'Notes']]
-    db.execute(u'''
-    select
-        date(e.start_time, 'unixepoch', 'localtime') as day,
-        ifnull(sum(ifnull(e.end_time, strftime('%s', 'now')) -
-                   e.start_time), 0) as day_total
-    from
-        entry e
-    where
-        e.sheet = ?
-    group by
-        day
-    order by
-        day asc;
-    ''', (sheet,))
-    days = db.fetchall()
-    days_iter = iter(days)
-    db.execute(u'''
-    select
-        date(e.start_time, 'unixepoch', 'localtime') as day,
-        e.start_time as start,
-        e.end_time as end,
-        ifnull(e.end_time, strftime('%s', 'now')) - e.start_time as
-            duration,
-        ifnull(e.description, '') as description
-    from
-        entry e
-    where
-        e.sheet = ?
-    order by
-        day asc;
-    ''', (sheet,))
-    entries = db.fetchall()
-    for i, (day, start, end, duration, description) in \
-            enumerate(entries):
-        date = displ_date(start)
-        diff = displ_total(duration)
-        if end is None:
-            trange = '%s -' % displ_time(start)
-        else:
-            trange = '%s - %s' % (displ_time(start), displ_time(end))
-        if last_day == day:
-            # If this row doesn't represent the first entry fo the
-            # day, don't display anything in the day column.
-            table.append(['', trange, diff, description])
-        else:
-            if last_day is not None:
-                # Use day_total set (below) from the previous
-                # iteration. This is skipped the first iteration,
-                # since last_day is None.
-                table.append(['', '', displ_total(day_total), ''])
-            cur_day, day_total = days_iter.next()
-            assert day == cur_day
-            table.append([date, trange, diff, description])
-            last_day = day
-
-    db.execute(u'''
-    select
-        ifnull(sum(ifnull(e.end_time, strftime('%s', 'now')) -
-                   e.start_time), 0) as total
-    from
-        entry e
-    where
-        e.sheet = ?;
-    ''', (sheet,))
-    total = displ_total(db.fetchone()[0])
-    table += [['', '', displ_total(day_total), ''],
-              ['Total', '', total, '']]
-    cmdutil.pprint_table(table, footer_row=True)
+    subprocess.call(('sqlite3', db.path))
 
 @command('start the timer for the current timesheet', name='in',
          aliases=('start',))
@@ -198,22 +104,32 @@ Notes may be specified for this period. This is exactly equivalent to
     parser.add_option('-s', '--switch', dest='switch', type='string',
                       help='Switch to another timesheet before \
 starting the timer.')
+    parser.add_option('-o', '--out', dest='out', action='store_true',
+                      default=False, help='''Clocks out before clocking \
+in''')
+    parser.add_option('-a', '--at', dest='at', type='string',
+                      help='''Set time of clock-in''')
     opts, args = parser.parse_args(args=args)
-    now = int(time.time())
     if opts.switch:
         sheet = opts.switch
         switch(db, [sheet])
     else:
         sheet = dbutil.get_current_sheet(db)
+    timestamp = cmdutil.parse_date_time_or_now(opts.at)
+    if opts.out:
+        clock_out(db, timestamp=timestamp)
     running = dbutil.get_active_info(db, sheet)
     if running is not None:
         raise SystemExit, 'error: timesheet already active'
+    most_recent_clockout = dbutil.get_most_recent_clockout(db, sheet)
+    if most_recent_clockout and timestamp < most_recent_clockout:
+        raise SystemExit, 'error: time periods could end up overlapping'
     description = u' '.join(args) or None
     db.execute(u'''
     insert into entry (
         sheet, start_time, description, extra
     ) values (?,?,?,?)
-    ''', (sheet, now, description, extra))
+    ''', (sheet, timestamp, description, extra))
 
 @command('delete a timesheet', aliases=('delete',))
 def kill(db, args):
@@ -357,14 +273,23 @@ Stop the timer for the current timesheet. Must be called after in.''')
     parser.add_option('-v', '--verbose', dest='verbose',
                       action='store_true', help='Show the duration of \
 the period that the out command ends.')
+    parser.add_option('-a', '--at', dest='at',
+                      help='Set time of clock-out')
     opts, args = parser.parse_args(args=args)
-    now = int(time.time())
-    active = dbutil.get_current_active_info(db)
+    clock_out(db, opts.at, opts.verbose)
+
+def clock_out(db, at=None, verbose=False, timestamp=None):
+    if not timestamp:
+        timestamp = cmdutil.parse_date_time_or_now(at)
+    active = dbutil.get_current_start_time(db)
     if active is None:
         raise SystemExit, 'error: timesheet not active'
-    active_id, active_time = active
-    if opts.verbose:
+    active_id, start_time = active
+    active_time = timestamp - start_time
+    if verbose:
         print timedelta(seconds=active_time)
+    if active_time < 0:
+        raise SystemExit, "Error: Negative active time"
     db.execute(u'''
     update
         entry
@@ -372,7 +297,7 @@ the period that the out command ends.')
         end_time = ?
     where
         entry.id = ?
-    ''', (now, active_id))
+    ''', (timestamp, active_id))
 
 @command('alter the description of the active period', aliases=('write',))
 def alter(db, args):
@@ -456,14 +381,16 @@ usage, see "%(prog)s --help".' % {'prog': os.path.basename(sys.argv[0])}
             active = duration
     print '%s: %s' % (sheet, active)
 
-@command('export a sheet to csv format', aliases=('csv', 'export'))
-def format(db, args):
-    import csv
+@command('display timesheet, by default the current one',
+         aliases=('export', 'format', 'show'))
+def display(db, args):
+    # arguments
+    parser = OptionParser(usage='''usage: %prog display [TIMESHEET]
 
-    parser = OptionParser(usage='''usage: %prog format [TIMESHEET]
-
-Export the current sheet as a comma-separated value format spreadsheet.
-If the final entry is active, it is ignored.
+Display the data from a timesheet in the range of dates specified, either
+in the normal timebook fashion (using --format=plain) or as
+comma-separated value format spreadsheet (using --format=csv), which
+ignores the final entry if active.
 
 If a specific timesheet is given, display the same information for that
 timesheet instead.''')
@@ -475,22 +402,38 @@ YYYY-MM-DD.')
                       metavar='DATE', help='Show only entries \
 ending before 00:00 on this date. The date should be of the format \
 YYYY-MM-DD.')
+    parser.add_option('-f', '--format', dest='format', type='string',
+                  default='plain',
+                  help="Select whether to output in the normal timebook \
+style (--format=plain) or csv --format=csv")
     opts, args = parser.parse_args(args=args)
+
+    # grab correct sheet
     if args:
         sheet = cmdutil.complete(dbutil.get_sheet_names(db), args[0],
                                  'timesheet')
     else:
         sheet = dbutil.get_current_sheet(db)
-    fmt = '%Y-%m-%d'
+
+    #calculate "where"
     where = ''
+    fmt = '%Y-%m-%d'
     if opts.start is not None:
-        start_date = datetime.strptime(opts.start, fmt)
-        start = cmdutil.datetime_to_int(start_date)
+        start = cmdutil.parse_date_time(opts.start)
         where += ' and start_time >= %s' % start
     if opts.end is not None:
-        end_date = datetime.strptime(opts.end, fmt)
-        end = cmdutil.datetime_to_int(end_date)
+        end = cmdutil.parse_date_time(opts.end)
         where += ' and end_time <= %s' % end
+    if opts.format == 'plain':
+        format_timebook(db, sheet, where)
+    elif opts.format == 'csv':
+        format_csv(db, sheet, where)
+    else:
+        raise SystemExit, 'Invalid format: %s' % opts.format
+
+def format_csv(db, sheet, where):
+    import csv
+
     writer = csv.writer(sys.stdout)
     writer.writerow(('Start', 'End', 'Length', 'Description'))
     db.execute(u'''
@@ -513,3 +456,88 @@ YYYY-MM-DD.')
         format(row[0]), format(row[1]), row[2], row[3]), rows))
     total_formula = '=SUM(C2:C%d)/3600' % (len(rows) + 1)
     writer.writerow(('Total', '', total_formula, ''))
+
+def format_timebook(db, sheet, where):
+    db.execute(u'''
+    select count(*) > 0 from entry where sheet = ?%s
+    ''' % where, (sheet,))
+    if not db.fetchone()[0]:
+        print '(empty)'
+        return
+
+    displ_time = lambda t: time.strftime('%H:%M:%S', time.localtime(t))
+    displ_date = lambda t: time.strftime('%b %d, %Y',
+                                         time.localtime(t))
+    displ_total = lambda t: \
+            cmdutil.timedelta_hms_display(timedelta(seconds=t))
+
+    last_day = None
+    table = [['Day', 'Start      End', 'Duration', 'Notes']]
+    db.execute(u'''
+    select
+        date(e.start_time, 'unixepoch', 'localtime') as day,
+        ifnull(sum(ifnull(e.end_time, strftime('%%s', 'now')) -
+                   e.start_time), 0) as day_total
+    from
+        entry e
+    where
+        e.sheet = ?%s
+    group by
+        day
+    order by
+        day asc;
+    ''' % where, (sheet,))
+    days = db.fetchall()
+    days_iter = iter(days)
+    db.execute(u'''
+    select
+        date(e.start_time, 'unixepoch', 'localtime') as day,
+        e.start_time as start,
+        e.end_time as end,
+        ifnull(e.end_time, strftime('%%s', 'now')) - e.start_time as
+            duration,
+        ifnull(e.description, '') as description
+    from
+        entry e
+    where
+        e.sheet = ?%s
+    order by
+        day asc;
+    ''' % where, (sheet,))
+    entries = db.fetchall()
+    for i, (day, start, end, duration, description) in \
+            enumerate(entries):
+        date = displ_date(start)
+        diff = displ_total(duration)
+        if end is None:
+            trange = '%s -' % displ_time(start)
+        else:
+            trange = '%s - %s' % (displ_time(start), displ_time(end))
+        if last_day == day:
+            # If this row doesn't represent the first entry of the
+            # day, don't display anything in the day column.
+            table.append(['', trange, diff, description])
+        else:
+            if last_day is not None:
+                # Use day_total set (below) from the previous
+                # iteration. This is skipped the first iteration,
+                # since last_day is None.
+                table.append(['', '', displ_total(day_total), ''])
+            cur_day, day_total = days_iter.next()
+            assert day == cur_day
+            table.append([date, trange, diff, description])
+            last_day = day
+
+    db.execute(u'''
+    select
+        ifnull(sum(ifnull(e.end_time, strftime('%%s', 'now')) -
+                   e.start_time), 0) as total
+    from
+        entry e
+    where
+        e.sheet = ?%s;
+    ''' % where, (sheet,))
+    total = displ_total(db.fetchone()[0])
+    table += [['', '', displ_total(day_total), ''],
+              ['Total', '', total, '']]
+    cmdutil.pprint_table(table, footer_row=True)
