@@ -71,6 +71,23 @@ def get_current_active_info(db):
     ''')
     return db.fetchone()
 
+def get_current_start_time(db):
+    db.execute(u'''
+    select
+        entry.id,
+        entry.start_time
+    from
+        entry
+    inner join
+        meta
+    on
+        meta.key = 'current_sheet' and
+        meta.value = entry.sheet
+    where
+        entry.end_time is null
+    ''')
+    return db.fetchone()
+
 def get_entry_count(db):
     db.execute(u'''
     select
@@ -85,3 +102,21 @@ def get_entry_count(db):
         sheet = m.value;
     ''')
     return db.fetchone()[0]
+
+def get_most_recent_clockout(db, sheet):
+    db.execute(u'''
+    select
+        end_time
+    from
+        entry
+    where
+        sheet = ?
+    order by
+        -end_time
+    ''', (sheet,))
+    row = db.fetchone()
+    if not row:
+        # we've never clocked out on this timesheet
+        return None
+    else:
+        return row[0]
